/***************************************************************************************************
 Project : RS485 klient pro mereni teploty a indikaci stavu
 Date    : 11/2014
 Author  : Jan Matuska, matuska@starhill.org                          
 Company : www.starhill.org
 Obsah   : obsahuje softwarovy UART a komunikaci s cidly DS1820 (DS18B20) dle id zarizeni
 Popis   : http://walda.starhill.org/elektro/rs485teplomer/
 ---------------------------------------------------------------------------------------------------
 Chip type                : ATTiny85
 Clock frequency          : 1.0 MHz (internal RC)
 Compiler                 : avr-gcc 
 Fuses                    : hFuse=0xDF, lFuse=x062, eFuse=0xFF
                            hFuse ma nastavenou log.0 na SPIEN, EESAVE
                            lFuse ma nastavenou log.0 na CKDIV8 SUT0 CKSEL3 CKSEL2 CKSEL0
 ---------------------------------------------------------------------------------------------------
 Komunikace: 1200bps, 1 stop bit, bez parity,

 Formt paketu:  *  d  s [fc] #   (vse ASCII znaky, velke, p. *AMZ#)
                 * = uvodni znak paketu
                 d = adresa prijemce (interval A-Z) (=lokalni adresa u prijmu)
                 s = adresa odesilatele (interval A-Z)                 
                 [fc] = zadny nebo vice znaku jako argumenty (interval 0-9A-F, u TX i | a -)
                 # = znak konce paketu
 ---------------------------------------------------------------------------------------------------
 Pkaz:  *dsZ#   vrati prumerovanou teplotu teplomeru a zapne LED                   
          *dsV#   vrati prumerovanou teplotu teplomeru a vypne LED
          *dsR#   vrati prumerovanou teplotu teplomeru a zapne rychle blikani LED
          *dsP#   vrati prumerovanou teplotu teplomeru a zapne pomale blikani LED
          *ds?#   vrati prumerovanou teplotu teplomeru, ?=libovolny znak mimo ZVRP          
          *ds#    vrati prumerovanou teplotu teplomeru
 Pklad: *XM#  *XMZ#  *XMV#  apod.          
 Odpov: * d s led [t2 t1 t0] #     (p. *MAV4F13EA#)
          d   = adresa prijemce = S z prijateho paketu
          s   = adresa odesilatele = lokalni adresa
          led = stav LED Z,V,R,P
          [t2 t1 t0] = data pro nastaveny teplomer (tzn. trojice vzdy pro kazdy nastaveny teplomer)
                t2 t1 t0 = tri pulbyte integeru teploty C, p. 3E3 = 0x03E3 = 995 = -0.5C                  
***************************************************************************************************/
#define F_CPU 1000000UL  // 1 MHz
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include "suart.c"
#include "onewire.c"


// Promenne v EEPROM  ------------------------------------------------------------------------------
#define PCIDEL 5   // max pocet teplomeru
#define EEPROM_SECTION __attribute__ ((section (".eeprom")))
typedef struct{                                                                                        
  unsigned char   id[8];
  unsigned char   type;
} EECID;
EECID ee_cidla[PCIDEL] EEPROM_SECTION = {
  {.type=0},
  {.type=0},
  {.type=0},
  {.type=0},
  {.type=0}
};
unsigned char ee_myaddr EEPROM_SECTION = 'X';
signed char ee_osc = 0;
// /Promenne v EEPROM  -----------------------------------------------------------------------------


// Ovladani indikacni LED --------------------------------------------------------------------------
#define LED_OFF    TCCR1 = 0; GTCCR=0; PORTB &= ~(1<<PB4);
#define LED_ON     TCCR1 = 0; GTCCR=0; PORTB |= (1<<PB4);
#define LED_SLOW   TCCR1 = 0x0D; GTCCR|=(1<<COM1B1)|(1<<PWM1B);
#define LED_FAST   TCCR1 = 0x0B; GTCCR|=(1<<COM1B1)|(1<<PWM1B);

//GTCCR|=(1<<COM1B1)|(1<<PWM1B);

#define IS_LED_ON   !(PORTB&(1<<PB4))
#define OFF         'V'
#define ON          'Z'
#define SLOW_BLINK  'P'
#define FAST_BLINK  'R'
// /Ovladani indikacni LED, sviti="topi" do dane smycky --------------------------------------------


// globani promenne
volatile unsigned char fcled = 'V';                 // rezim LED
volatile unsigned char myaddr = 'X';                // lokalni adresa zarizeni
struct{                                                                                       
  unsigned int    data[4];
  unsigned int    outval; 
} cidla[PCIDEL];                                    // struktura pro prumerovani teplot

                                        
                              
                              

// -------------------------------------------------------------------------------------------------
// Pevede cislo 0-15 na znak 0-9A-F
// Vstup:  cislo, z nej se berou jen spodni 4 bity
// Vystup: znak 0-9A-F
// -------------------------------------------------------------------------------------------------
unsigned char suartNibbleToZnak(unsigned char nibble) {
  unsigned char retval = '0'+(nibble&0x0F);
  if(retval>=58) retval+=7;
  return(retval);
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// Pevede znak 0-9A-F na cislo 0-15
// Vstup:  znak 0-9A-F
// Vystup: cislo 0-15, i za predpokladu, ze vstup je mimo povoleny interval
// -------------------------------------------------------------------------------------------------
unsigned char suartZnakToNibble(unsigned char znak) {
  if (znak>='A') znak=znak-7;
  return((znak-'0')&0x0F);
}
// -------------------------------------------------------------------------------------------------



// -------------------------------------------------------------------------------------------------
// Provede zmereni teploty nastavenych teplomeru a zprumerovani za posledni 4 mereni
// Vstup: nic, resp. eeprom data s nastavenymi cidly
// Vystup: naplni obsah struktury cidla                                                                              
// -------------------------------------------------------------------------------------------------
void mereni() {
  unsigned int tmp;
  unsigned char i, bufferOK, cidlo;  
  // -- prevod teploty v cidlech -- 
  cidlo=PCIDEL;
  while(cidlo--) {
    if(suartZnakToNibble(eeprom_read_byte(&ee_cidla[cidlo].type))>0) {
      i=8;
      while(i--) ow_device_rom[i] = eeprom_read_byte(&ee_cidla[cidlo].id[i]);
      ow_execute_convert();
    }
  }
  _delay_ms(750);
  // -- vycteni teploty  -- 
  cidlo=PCIDEL;
  while(cidlo--) {
    if(suartZnakToNibble(eeprom_read_byte(&ee_cidla[cidlo].type))>0) {
      tmp=0;
      i=3;
      bufferOK=TRUE; 
      while(i--) {
        cidla[cidlo].data[i+1] = cidla[cidlo].data[i];
        tmp += cidla[cidlo].data[i];
        if(cidla[cidlo].data[i]==0) bufferOK=FALSE;  
      }
      i=8;
      while(i--) ow_device_rom[i] = eeprom_read_byte(&ee_cidla[cidlo].id[i]);
      cidla[cidlo].data[0] = ow_get_temp(eeprom_read_byte(&ee_cidla[cidlo].type), READ_ONLY);
      tmp += cidla[cidlo].data[0];
      if( (bufferOK==FALSE) || (cidla[cidlo].data[0]==0) ) {
        cidla[cidlo].outval = cidla[cidlo].data[0];
        i=4;
        while(i--) cidla[cidlo].data[i]=cidla[cidlo].outval;
      } else {
        cidla[cidlo].outval = tmp >> 2;
      }      
    }
  }
}  
// -------------------------------------------------------------------------------------------------





// Startpoint
int main(void) {

  unsigned int tmp;
  unsigned char i,j,k;
  signed char osc;

  // kalibrace oscilatoru
  osc = eeprom_read_byte(&ee_osc);
  if((osc>-50) && (osc<50)) OSCCAL += osc; 

  // Nastaveni Timer1 - nastaveni rezimu PWM pro blikani LED, ostatni v makrech LED_xxx
  OCR1B=0x080;
  OCR1C=0x0FF;
  LED_OFF;
  
  // nastaveni suartu
  suartInit(); 
  
  // globani povoleni preruseni
  sei(); 

  // nacteni adresy zarizeni z EEPROM
  myaddr = eeprom_read_byte(&ee_myaddr);
  if(myaddr<'A'||myaddr>'Z') myaddr='X';
  
  // nastaveni vystupnich pinu
  DDRB |= (1<<DDB3)|(1<<DDB4);

  // zapnuti psa  
  wdt_enable(WDTO_8S);	
  
  // vychozi naplneni prumerovacich bufferu
  mereni();

  // mainloop  
  while(1) {
  
  
    // krokovani
    wdt_reset();
    _delay_ms(10); 
  
  
    // detekce paketu pro toto zarizeni
    if ((suartState==STBY)&&(suartPacket[0]=='*')&&(suartPacket[1]==myaddr)) {
   
      // zneplatneni paketu
      suartPacket[0]='%';  
   
      // detekce noveho paketu na poslani teplot
      if( (suartPacket[4]=='#') || (suartPacket[3]=='#') ) {
         switch(suartPacket[3]) {
          case 'V':
            fcled = OFF;
            LED_OFF;
            break;          
          case 'Z':
            fcled = ON;
            LED_ON;
            break;          
          case 'P':
            fcled = SLOW_BLINK;
            LED_SLOW;
            break;          
          case 'R':
            fcled = FAST_BLINK;
            LED_FAST;
            break;          
        }
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendChar(fcled);
        i=0;
        while(i<PCIDEL) {
          if(suartZnakToNibble(eeprom_read_byte(&ee_cidla[i].type))>0) {
            tmp = cidla[i].outval;
            suartSendChar(suartNibbleToZnak((unsigned char) ((tmp>>8) & 0x0F)));      
            suartSendChar(suartNibbleToZnak((unsigned char) ((tmp>>4) & 0x0F)));      
            suartSendChar(suartNibbleToZnak((unsigned char) (tmp & 0x0F)));        
          }
          i++;
        }
        suartSendPacketEOP();
        mereni();
      } 
      
      
      // [*DSSAx#] paket na nastaveni adresy zarizeni
      if( (suartPacket[6]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='A') ) {
        if( (suartPacket[5]>='A') && (suartPacket[5]<='Z') ) {
          eeprom_write_byte(&ee_myaddr,suartPacket[5]);
          myaddr = eeprom_read_byte(&ee_myaddr);
        }
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendPacketEOP();      
      }


      // [*DSSS#] provede prohledani onewire sbernice
      if( (suartPacket[5]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='S') ) {
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendChar('|');
        ow_search_reset();
        while(ow_search_find()) {
          i=0;
          while(i<8) {
            suartSendChar(suartNibbleToZnak( (ow_device_rom[i]>>4) & 0x0F) );
            suartSendChar(suartNibbleToZnak( ow_device_rom[i] & 0x0F) );
            i++;
          }
          suartSendChar('|');
        }             
        suartSendPacketEOP();      
      }
      
      
      // [*DSSR#] vypis nastaveni teplomeru
      if( (suartPacket[5]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='R')) {
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendChar('|');
        i=0;
        while(i<PCIDEL) {
          suartSendChar(suartNibbleToZnak(i));
          suartSendChar('-');
          k=eeprom_read_byte(&ee_cidla[i].type);
          suartSendChar(suartNibbleToZnak(k));
          suartSendChar('-');
          j=0;
          while(j<8) {
            k=eeprom_read_byte(&ee_cidla[i].id[j]);
            suartSendChar(suartNibbleToZnak( (k>>4) & 0x0F) );
            suartSendChar(suartNibbleToZnak( k & 0x0F) );
            j++;
          } 
          suartSendChar('|');
          i++; 
        }
        suartSendPacketEOP();      
      }
      
      
      //*DSSCctssssssssssssssss#  prida do seznamu id teplomeru
      if((suartPacket[23]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='C')) {
        i = suartZnakToNibble(suartPacket[5]);
        j = suartZnakToNibble(suartPacket[6]);
        if( (i<PCIDEL) && (j<3) ) {
          suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
          suartSendChar('|');
          suartSendChar(suartNibbleToZnak(i));
          suartSendChar('-');
          suartSendChar(suartNibbleToZnak(j));
          suartSendChar('-');
          eeprom_write_byte(&ee_cidla[i].type, j);
          k=0;
          while(k<8) {
            j =  suartZnakToNibble(suartPacket[(7+(k<<1))])<<4;
            j |= suartZnakToNibble(suartPacket[(8+(k<<1))]);
            eeprom_write_byte(&ee_cidla[i].id[k], j);
            suartSendChar(suartNibbleToZnak( (j>>4) & 0x0F) );
            suartSendChar(suartNibbleToZnak( j & 0x0F) );                     
            k++;
          }
          suartSendChar('|');
          suartSendPacketEOP();   
        }
      }      


      // [*DSSDc#] odstraneni nastaveni cidla
      if( (suartPacket[6]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='D') ) {
        i = suartZnakToNibble(suartPacket[5]);
        if( (i<PCIDEL) ) {
          eeprom_write_byte(&ee_cidla[i].type, 0);
          suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
          suartSendChar('|');
          suartSendChar(suartNibbleToZnak(i));
          suartSendChar('-');
          suartSendChar('0');                  
          suartSendChar('|');
          suartSendPacketEOP();   
        }
      }
      
      // [*DSSTx#] paket na spusteni testu casovani
      if( (suartPacket[6]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='T') ) {
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendChar('T');
        suartSendChar(suartPacket[5]);
        suartSendPacketEOP();         
        if(suartPacket[5]=='0') suartTestTiming(0); 
        if(suartPacket[5]=='1') suartTestTiming(1);
        while(1); 
      }

      // [*DSSOx#] paket na nastaveni kalibrace
      if( (suartPacket[6]=='#')&&(suartPacket[3]=='S')&&(suartPacket[4]=='O') ) {
        suartSendPacketHeader(suartPacket[SRC_ADDR], myaddr);
        suartSendChar('O');
        if(suartPacket[5]=='1') {
          suartSendChar('+');
          osc++;
          OSCCAL++;
          eeprom_write_byte(&ee_osc, osc);
          suartTestTiming(0);
        } 
        if(suartPacket[5]=='0') {
          suartSendChar('-');
          osc--;
          OSCCAL--;
          eeprom_write_byte(&ee_osc, osc);
          suartTestTiming(0);
        } 
        suartSendPacketEOP();         
        while(1); 
      }
            
    }
    // /detekce paketu pro toto zarizeni
    
    
    
    // detekce broadcast paketu na identifikaci
    if ((suartState==STBY)&&(suartPacket[0]=='*')&&(suartPacket[1]=='?')) {
      // zneplatneni paketu a zaloha zdrojove adresy
      suartPacket[0]='%';  
      j = suartPacket[SRC_ADDR];
      // pocka dobu dle adresy
      i='A'-1;
      while((myaddr!=i) && (i<='Z')) {
        _delay_ms(1000);
        i++;
        if(myaddr==i) {
          suartSendPacketHeader(j, myaddr);
          suartSendPacketEOP();        
        }
        wdt_reset();
      }
    } 
    // /detekce broadcast paketu na identifikaci       
  
  
  }
  // /mainloop

}



