// -------------------------------------------------------------------------------------------------
// Modul:     ONEWIRE
// Funkce:    Komunikace s cidly DS1820 a DS18B20, obsahuje vyhledavaci algoritmus
// Pozadavky: 1 GPIO pro one-wire sbernici
// -------------------------------------------------------------------------------------------------


// nastaveni portu a pinu pro sbernici  
#define TX          DDRB |= (1<<PB3); PORTB &= ~(1<<PB3)
#define RX          PORTB |= (1<<PB3); DDRB &= ~(1<<PB3);
#define RXPIN       PINB&(1<<PB3)
// indexy
#define NOSENSOR          0
#define DS1820            1
#define DS18B20           2
#define TRUE              1
#define FALSE             0
#define CONVERT_AND_READ  0
#define READ_ONLY         1

// globalni promenne
static unsigned char ow_last_discrepancy;            // pro vyhledavani OW
static unsigned char ow_last_device_flag;            // pro vyhledavani OW - vyhledavani ukonceno
unsigned char ow_device_rom[8];                      // pole pro ulozeni ROM OW zarizeni


// prototypy
static unsigned char ow_detect_presence(void);
static void ow_write_one(void);
static void ow_write_zero(void);
static unsigned char ow_read_bit(void);
static void ow_write_byte(unsigned char tosend);
static unsigned char ow_read_byte(void);
void ow_search_reset(void);
unsigned char ow_search_find(void);
void ow_execute_convert();
unsigned int ow_get_temp(unsigned char type_of_sensor, unsigned char mode);



// -------------------------------------------------------------------------------------------------
// [PRIVATE] Funkce na zapis a cteni na sbernici OneWire
// -------------------------------------------------------------------------------------------------
// provede reset a test prezence na OW sbernici 
static unsigned char ow_detect_presence(void) {
  unsigned char out=1;              // vychozi navratova hodnota
  RX;                               // vychozi stav sbernice
  _delay_us(1000);                  // pro ustaleni
  TX;                               // bus low
  _delay_us(480);                   // cas pro prikaz reset
  RX;                               // uvolneni sbernice
  _delay_us(70);                    // cekani na potvrzeni teplomerem
  if(RXPIN) out=0;                  // pokud detekovana log.1, tak teplomer na sbernici neni
  _delay_us(410);                   // pauza pred dalsi komunikaci
  return out;                       // vrati stav 1=zarizeni nalezeno,  0=zarizeni nenalezeno
}
// posle na sbernici log.1
static void ow_write_one(void) {
  TX;                               // bus low
  _delay_us(6);                     // pauza definujici log.1
  RX;                               // uvolneni sbernice
  _delay_us(64);                    // pauza pred dalsi komunikaci 
}
// posle na sbernici log.0
static void ow_write_zero(void) {
  TX;                               // bus low
  _delay_us(60);                    // pauza definujici log.0
  RX;                               // uvolneni sbernice
  _delay_us(10);                    // pauza pred dalsi komunikaci 
}
// precte jeden bit ze sbernice
static unsigned char ow_read_bit(void) {
  unsigned char out=0;              // vychozi navratova hodnota bitu
  TX;                               // bus low
  _delay_us(6);                     // pauza pro stav cteni
  RX;                               // uvolneni sbernice
  _delay_us(9);                     // pauza pro reakci teplomeru
  if(RXPIN) out=1;                  // test stavu sbernice, vlastni cteni
  _delay_us(55);                    // pauza pred dalsi komunikaci  
  return out;                       // prectena hodnota, 1 nebo 0
}
// odeslne na sbernici jeden byte. Odesila se prvni LSB
static void ow_write_byte(unsigned char tosend) {
  int n=8;
  while(n--) {
    if(tosend&1) ow_write_one(); else ow_write_zero();
    tosend >>= 1;
  }
}
// prijde ze sbernice jeden byte. Prijima jako prvni LSB.
static unsigned char ow_read_byte(void) {
  int n=0, out=0;
  while(n<8) {
    if(ow_read_bit()) out |= (1<<n);
    n++;        
  }
  return out;
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Reset vyhledavani - bude nalezeno opet prvni zarizeni 
// -------------------------------------------------------------------------------------------------
void ow_search_reset(void) {
  ow_last_discrepancy = 0;
  ow_last_device_flag = FALSE;
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Vyhleda zarizeni na onewire sbernici, funkce se vola opakovane, dokud nevrati FALSE 
// -------------------------------------------------------------------------------------------------
// Vystup:  TRUE  - nalezeno zarizeni, adresa je v poli ow_device_rom
//          FALSE - nebylo nalezeno zadne nebo dalsi zarizeni 
// -------------------------------------------------------------------------------------------------
unsigned char ow_search_find(void)
{
  unsigned char id_bit_number, last_zero, rom_byte_number, search_result;
  unsigned char id_bit, cmp_id_bit;
  unsigned char rom_byte_mask, search_direction=0;
  // inicializace provoznich promennych
  id_bit_number = 1;
  last_zero = 0;
  rom_byte_number = 0;
  rom_byte_mask = 1;
  search_result = 0;
  // pokud jiz neni co hledat
  if(ow_last_device_flag) return FALSE;
  // reset a test pritomnosti zarizeni na sbernici
  cli();
  if( !ow_detect_presence() ) {
    ow_search_reset();
    sei();
    return FALSE;
  }
  // posle prikaz search
  ow_write_byte(0xF0);   
  // smycka jdouci po bitech serioveho cisla zarizeni
  while(id_bit_number<65) {
    // precte ze zarizeni bit a jeho negaci
    id_bit = ow_read_bit();
    cmp_id_bit = ow_read_bit();
    // na sbernici neni zadne zarizeni
    if( (id_bit==1) && (cmp_id_bit==1) ) break;  
    // na sbernici jsou zarizeni, ktere maji na teto pozici stejnou hodnotu bitu serioveho cisla 
    if(id_bit!=cmp_id_bit) search_direction=id_bit; 
    // na sbernici doslo ke kolizi - zarizeni maji na teto pozici rozdilne hodnoty bitu 
    if(!id_bit && !cmp_id_bit) {
      // pokracujeme nulou
      if(id_bit_number>ow_last_discrepancy) { 
        last_zero=id_bit_number; 
        search_direction=0; 
      }
      // pokracujeme jednickou
      if(id_bit_number==ow_last_discrepancy) search_direction=1;
      // pokracujeme tim, co jiz bylo drive zjisteno
      if(id_bit_number<ow_last_discrepancy) 
        search_direction=ow_device_rom[rom_byte_number] & rom_byte_mask;
    }
    // zapis bitu do ROM_NO a odeslani na sbernici pro vyber smeru prohledavani
    if(search_direction==1) {
      ow_device_rom[rom_byte_number] |= rom_byte_mask;
      ow_write_one();
    } else {
      ow_device_rom[rom_byte_number] &= ~rom_byte_mask;
      ow_write_zero();
    }
    // inkrementace citace projitych bitu
    id_bit_number++;
    // posun masky v ramci byte v poli ROM_NO 
    rom_byte_mask<<=1;
    // pokud je maska 0, tak se bude zapisovat do dalsiho byte v poli ROM_NO
    if(rom_byte_mask==0) {
      rom_byte_number++;
      rom_byte_mask=1;
    }        
  }
  // prohledano vsech 64 bitu    
  if(id_bit_number>64) {
    ow_last_discrepancy = last_zero;
    if(ow_last_discrepancy==0) ow_last_device_flag=TRUE;
    search_result=TRUE;
  }    
  // pokud nebylo hledani uspesne, tak reset stavu hledani
  if( !search_result || !ow_device_rom[0] ) {
    ow_search_reset();
    search_result=FALSE;
  }
  // vrati vysledek hledani
  sei();
  return search_result;  
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Spusti prevod teploty v cidle s id v ow_device_rom
// -------------------------------------------------------------------------------------------------
void ow_execute_convert() {
  unsigned char i;
  cli();
  // -- prikaz pro zmereni teploty v cidle --
  ow_detect_presence();
  //ow_write_byte(0x0CC);
  ow_write_byte(0x055); 
  i=0;
  while(i<=7) ow_write_byte(ow_device_rom[i++]);
  ow_write_byte(0x44);
  sei();
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Precte teplotu z teplomeru daneho typu v cidle s id v ow_device_rom
//           Nacte teplotu z teplomeru a vrati ji ve formatu 1000+t*10
//           Priklad: 23.5C = 1235,  -10.5C = 895
// -------------------------------------------------------------------------------------------------
unsigned int ow_get_temp(unsigned char type_of_sensor, unsigned char mode) {
  unsigned char data_lo, data_hi,i;
  signed int teplota = 0;
  unsigned int tmp = 0;
  // -- pripadne provedeni zmereni aktualni teploty v cidle -- 
  if(mode==CONVERT_AND_READ) {
    ow_execute_convert();
    _delay_ms(750);  
  }
  cli();
  // -- prikaz pro precteni teploty -- 
  ow_detect_presence();
  //ow_write_byte(0x0CC);
  ow_write_byte(0x055); 
  i=0;
  while(i<=7) ow_write_byte(ow_device_rom[i++]);
  ow_write_byte(0x0BE);
  data_lo=ow_read_byte();     // 1. byte scratchpadu teplomeru = spodni byte teploty
  data_hi=ow_read_byte();     // 2. byte scratchpadu teplomeru = horni byte teploty
  // -- zpracovani vysledku --
  if(type_of_sensor==DS18B20) teplota = ((data_hi<<8) | data_lo) * 0.625;
  if(type_of_sensor==DS1820) teplota = ((data_hi<<8) | data_lo) * 5; 
  tmp = teplota + 1000;
  sei();
  // -- kontrola teploty mezi -50C a 100C --
  if((tmp<500)||(tmp>2000)) { 
    return 0;
  } else {
    return tmp;
  }
}
// -------------------------------------------------------------------------------------------------

