// -------------------------------------------------------------------------------------------------
// Modul:     SUART
// Funkce:    Realizuje funkci softwaroveho uartu s podporou rizeni RX/TX na RS485
// Pozadavky: INT0, TIMER0, 3 GPIO s tim, ze na jednom (RXD) musi byt INT0, konstanty pro 1MHz
// -------------------------------------------------------------------------------------------------


// ovladani vystupu suartu TXD 
#define SUART_ZERO PORTB &= ~(1<<PB0);
#define SUART_ONE  PORTB |= (1<<PB0);
// pin pro RXD - MUSI mit INT0 preruseni
#define SUART_RXPIN PINB&(1<<PB2)
// ovladani RE/DE na RS485
#define TX_MODE  PORTB |= (1<<PB1);
#define RX_MODE  PORTB &= ~(1<<PB1);
// nastaveni I/O pinu, TXD a RE/DE vystup, RXD vstup s INT0
#define SUART_IO_OUTPUT DDRB|=(1<<DDB0)|(1<<DDB1);
#define SUART_IO_INPUT  DDRB&=~(1<<DDB2);
// rezimy SUARTu
#define STBY   0
#define SENDCHAR 20
#define RECEIVE 21
#define TIMING 30
// ovladani casovace - bitrate 1200bps pro 1MHz clockrate
#define SUART_TICKS_BITRATE   OCR0A = 105;                                // rychlost 1200bps, T=833.33us
#define SUART_TICKS_STARTBIT  OCR0A = 153;                                // rychlost 800bps, 1.5*T=1.25ms
#define SUART_TICKS_ON        TCNT0 = 0; TCCR0B = 0x02;                   // spusti casovac
#define SUART_TICKS_OFF       TCCR0B = 0x00; TIFR |= (1<<OCF0A);          // zastavi casovac
// povoleni a zakazani detekce noveho startbitu - tzn. prijmu noveho byte
#define SUART_NEWFRAME_DETECT_DISABLE   GIMSK &= ~(1<<INT0);
#define SUART_NEWFRAME_DETECT_ENABLE    GIFR |= (1<<INTF0); GIMSK |= (1<<INT0);
// indexy
#define SRC_ADDR 2
#define DST_ADDR 1
// velikost bufferu
#define BUFFER_SIZE 32


// globani promenne
volatile unsigned int  suartFrame;                            // registr na uart ramec k odeslano
volatile unsigned char suartState = STBY;                     // vychozi stav SW uartu
volatile unsigned char suartPacket[BUFFER_SIZE];              // pole pro ulozeni paketu
volatile unsigned char suartPacketPointer=0;                  // ukazatel pozice v bufferu
volatile unsigned char suartBitCount;                         // pocitadlo bitu 


// prototypy
void suartSendChar(unsigned char znak);
void suartTestTiming(unsigned char typ);
void suartSendPacketHeader(uint8_t dst, uint8_t src);
void suartSendPacketEOP(void);
void suartInit(void);



// -------------------------------------------------------------------------------------------------
// [INT] Obsluha preruseni RX pinu - spusti sekvenci cteni 1 byte
//       Detekce sestupne hrany startbitu
// -------------------------------------------------------------------------------------------------
ISR(INT0_vect) {
  SUART_NEWFRAME_DETECT_DISABLE;                              // vypne dalsi preruseni na INT0
  SUART_TICKS_STARTBIT;                                       // nastavi 1.5 nasobek bitu
  SUART_TICKS_ON;                                             // spusti casovac
  suartState = RECEIVE;                                       // nastavi stav suartu na prijem ramce
  suartBitCount = 0;                                          // pocitadlo prijatych bitu uart ramce
  suartFrame = 0;                                             // prijaty ramec - data 
}

// -------------------------------------------------------------------------------------------------
// [INT] Obslusha preruseni o Timer0 a komprarace s OCR0A
// -------------------------------------------------------------------------------------------------
// Obslusha preruseni o Timer0 a komprarace s OCR0A 
ISR(TIMER0_COMPA_vect) {
  
  // rezim prijmu UART ramce z linky D0 D1 D2 D3 D4 D5 D6 D7 1
  if ( suartState==RECEIVE ) {

    SUART_TICKS_BITRATE;                                                 // nastaveni castovace na bitrate 

    if( suartBitCount<8 ) {                                              // prijem dat
      if( SUART_RXPIN ) {                                                // stav na vstupnim pinu
        suartFrame |= (1<<(suartBitCount));                              // zapis do registru dat
      }      
    } 
           
    if( suartBitCount==8 ) {                                             // prijem stopbitu (nekontrolujeme)
      if( suartFrame>=32 ) {                                             // ignorace ridicich znaku
        if( suartFrame=='*' ) suartPacketPointer = 0;
        if( suartPacketPointer>=BUFFER_SIZE ) suartPacketPointer=0; 
        suartPacket[suartPacketPointer] = (unsigned char) suartFrame;    // zapis do bufferu
        suartPacketPointer++;
      }
      SUART_NEWFRAME_DETECT_ENABLE;                                                                                                   
    }
    
    if( suartBitCount==12 ) {                                            // cas pro novy byte
      SUART_TICKS_OFF;                                                   // zastaveni casovace
      suartState = STBY;                                                 // uvedeni suart do klidoveho stavu                                                                               
    }
    
    suartBitCount++;                                                     // pocitadlo prijatych bitu    
    return;  
  }
  
  // rezim odesilani znaku
  if ( suartState==SENDCHAR ) {
    if (suartFrame>0) {                                                  // test obsahu ramce
      if( (suartFrame)&~(0xFFFE) ) {                                     // odeslani LSB ramce
        SUART_ONE;                                                       // Hi na sbernici
      } else {
        SUART_ZERO;                                                      // Lo na sbernici
      }          
      suartFrame>>=1;                                                    // posun MSB smerem k LSB
    } else {
      SUART_TICKS_OFF;                                                   // zastaveni casovace
      suartState = STBY;                                                 // uvedeni suartu do klidoveho stavu
    }
    return;
  }

  // rezim testu casovani 
  if ( suartState==TIMING ) {
    if (!suartFrame) {                                                   // test obsahu ridici promenne
        SUART_ONE;                                                       // Hi na sbernici
        suartFrame=1;                                                    
    } else {
        SUART_ZERO;                                                      // Lo na sbernici
        suartFrame=0;                                                    
    }          
    return;
  }
  
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Odesle znak na TXD pin 
//          suartFrame:  1 D7 D6 D5 D4 D3 D2 D1 D0 0   (vysilani zprava doleva)
// -------------------------------------------------------------------------------------------------
void suartSendChar(unsigned char znak) {
  while( suartState != STBY );
  suartFrame = 0x200 | (znak<<1);
  suartState = SENDCHAR;
  SUART_TICKS_BITRATE;
  SUART_TICKS_ON;
  while( suartState != STBY );
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Spusti test casovani - ukonceni resetem 
//          vstup:  0=zakladni bitrate, 1=1.5 nasobek periody zakladniho bitrate
// -------------------------------------------------------------------------------------------------
void suartTestTiming(unsigned char typ) {
  while( suartState != STBY );
  suartState = TIMING;
  if(typ) {
    SUART_TICKS_STARTBIT;
  } else {
    SUART_TICKS_BITRATE;
  }
  TX_MODE;  
  SUART_TICKS_ON;
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Odesle hlavicku paketu s prepnutim RS485 to TX rezimu
//          Vstup: dst - adresa prijemce, src - adresa odesilatele
// -------------------------------------------------------------------------------------------------
void suartSendPacketHeader(uint8_t dst, uint8_t src) {
  SUART_NEWFRAME_DETECT_DISABLE; 
  TX_MODE;   
  suartSendChar('*');
  suartSendChar(dst);
  suartSendChar(src);
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Odesle ukonceni paketu s naslednym prepnutim RS485 to RX rezimu
//          Vstup: dst - adresa prijemce, src - adresa odesilatele
// -------------------------------------------------------------------------------------------------
void suartSendPacketEOP(void) {
  suartSendChar('#');
  suartSendChar('\n');
  suartSendChar('\r');
  RX_MODE;
  SUART_NEWFRAME_DETECT_ENABLE;  
}
// -------------------------------------------------------------------------------------------------


// -------------------------------------------------------------------------------------------------
// [PUBLIC] Provede zakladni nastaveni SUARTu
// -------------------------------------------------------------------------------------------------
void suartInit(void) {
  // Timer0 do CTC rezimu - komparace s OCR0A
  TCCR0A |= (1<<WGM01);
  // Povoleni preruseni pri dosazeni timeru0 hodnoty OCR0A
  TIMSK |= (1<<OCIE0A);
  // Nastaveni externiho INT0 preruseni na sestupnou hranu  
  MCUCR |= (1<<ISC01);  
  // nastaveni I/O pinu
  SUART_IO_OUTPUT;
  SUART_IO_INPUT;
  // nastaveni linky do vychoziho stavu
  SUART_ONE;
  // Povoleni preruseni INT0 na sestupnou hranu - cekani na startbit 
  SUART_NEWFRAME_DETECT_ENABLE;
  // Vychozi nastaveni prijimace RS485
  RX_MODE;   
}
// -------------------------------------------------------------------------------------------------



